/*
 *  CES-10 bomber
 *  Este  um prottipo para o ltimo laboratrio do curso CES-10
 *  
 *  Autor: Carlos Henrique Q. Forster
 *  Instituto Tecnolgico de Aeronutica
 *  02/06/2008
 */

#include <graphics.h>
#include <math.h>
#include "poligono.h"

/* prototipos de algumas funcoes definidas nesse programa */

int intercept(int ax, int ay, int bx, int by, int cx, int cy, int dx, int dy,
              int *px, int *py, float *dist);
int inteceptpoli(int ax, int ay, int bx, int by, poligono p,
                 int *qx, int *qy, float *dist);
int intercept_cenario(int ax, int ay, int bx, int by, int *px, int *py, float *dist);
poligono explosao(int ax, int ay, int r);


/* sqr(x) retorna o quadrado de um float x */
float sqr(float x)
{
  return x*x;
}


/* intercept retorna 1 se dois segmentos de reta se interceptam ou 0 caso contrario
   segmento 1 eh (ax,ay)-(bx,by) e segmento 2 eh (cx,cy)-(dx,dy)
   em (px,py) eh colocado o ponto comum se houver interseccao dos segmentos
   e em dist eh colocada a distancia a partir do ponto (ax,ay)
*/
int intercept(int ax, int ay, int bx, int by, int cx, int cy, int dx, int dy,
              int *px, int *py, float *dist)
{
  float ux,uy,vx,vy, ab, cd, alfa, beta, det, det_u, det_v;
  
  ab=sqrt(sqr(bx-ax)+sqr(by-ay)); /*comprimento de a-b */
  cd=sqrt(sqr(dx-cx)+sqr(dy-cy));
  
  ux=(bx-ax)/ab; /* (ux,uy) vetor unitario na direcao b-a */
  vx=(dx-cx)/cd; /* (vx,vy) vetor unitario na direcao d-c */
  uy=(by-ay)/ab;
  vy=(dy-cy)/cd;
  
  det=-ux*vy+vx*uy;
  if (fabs(det) < 0.01) return 0; /*ignora o caso de quase paralelas*/
  
  det_u= -(cx-ax)*vy + (cy-ay)*vx;
  det_v= ux*(cy-ay) - uy*(cx-ax);
  
  alfa=det_u/det; /* a+alfa*u = b+beta*v solucao pela regra de Cramer*/
  beta=det_v/det;
  
  /* testa se houve de fato interseccao dos segmentos de reta */
  if (0<=alfa && alfa<=ab && 0<=beta && beta<=cd)
  {
    *px=(int)(ax+alfa*ux);
	*py=(int)(ay+alfa*uy);
	*dist=alfa;
	return 1;
  }
  else return 0;
}


/* interceptpoli retorna o numero de vezes que um segmento intercepta o
   contorno do poligono p.
   o segmento eh (ax,ay)-(bx,by)
   em (qx,qy) eh colocado o ponto de interseccao mais proximo de (ax,ay)
   em dist eh colocada essa distancia minima
 */
int interceptpoli(int ax, int ay, int bx, int by, poligono p,
                 int *qx, int *qy, float *dist)
{
  int i=0, px, py, pxclosest=30000, pyclosest=30000, count=0;
  float adist, mindist=100000.0F;
  vertice *v, *w;
  v=p.vert;
  w=v->prox;
  while(w) /* testa segmento contra cada lado do poligono */
  {
    i=intercept(ax, ay, bx, by, v->x, v->y, w->x, w->y, &px, &py, &adist);
	if(i)
	{
	  count++; /* conta o numero de cruzamentos */
	  if(adist<mindist) /* guarda o ponto de interseccao mais proximo de a*/
	  {
	    mindist=adist;
		pxclosest=px;
		pyclosest=py;
	  }
	}
	v=w;
	w=w->prox;
  }
  w=p.vert;
    i=intercept(ax, ay, bx, by, v->x, v->y, w->x, w->y, &px, &py, &adist);
	if(i)
	{
	  count++;
	  if(adist<mindist)
	  {
	    mindist=adist;
		pxclosest=px;
		pyclosest=py;
	  }
	}
  if(count>0)
  {
    *qx=px;
	*qy=py;
	*dist=mindist;
  }	   
  return count;
}


/* definicao do cenario (variaveis globais)
   consiste em um conjunto de poligonos que funcionam como paredes
   o vetor cenario pode ter ate 30 poligonos
   o numero de poligonos do cenario eh colocado em n_cenario
*/
poligono cenario[30];
int n_cenario=0;


/* intercept_cenario retorna 1 se o segmento intercepta algum poligono do cenario
   ou 0 caso contrario.
   o segmento de teste eh (ax,ay)-(bx,by)
   o ponto de interseccao mais proximo de (ax,ay) eh colocado em (px,py) e
   sua distancia a (ax,ay) eh colocada em dist
*/
int intercept_cenario(int ax, int ay, int bx, int by, int *px, int *py, float *dist)
{
  int i,qx,qy,k, changed=0, bestx, besty;
  float mindist=100000.0F, adist;
  for(i=0; i<n_cenario; i++) /*testa segmento contra cada poligono do cenario*/
  {
    k=interceptpoli(ax,ay,bx,by,cenario[i],&qx,&qy,&adist);
	if(k>0)
	{
	  if(adist<mindist) /*guarda o ponto de interseccao mais proximo de a*/
	  {
	    mindist=adist;
        bestx=qx;
		besty=qy;
		changed=1;
	  }
	}
  }
  if(changed)
  {
    *dist=mindist;
	*px=bestx;
	*py=besty;
  }
  return changed;
}


/* explosao retorna um poligono com centro em ax,ay e raio maximo r que se adapta aos
   elementos do cenario, isto eh, nao atravessa as paredes
 */
poligono explosao(int ax, int ay, int r)
{
  int i, bx, by, px, py;
  float s,c, dist, k;
  poligono p;
  p=novo(RED,YELLOW);
  for(i=0; i<30; i++)
  {
    s=sin(3.1416*2.0*i/30.0); /*poligono de 30 lados*/
    c=cos(3.1416*2.0*i/30.0);
	k=(i%2)?1.0:0.8; /*zig-zag*/
	bx=ax+k*r*c;
	by=ay+k*r*s;
	px=bx; /* se nao colidir com a parede, fica p=b */
	py=by;
	intercept_cenario(ax,ay,bx,by, &px, &py, &dist); /*colisao contra paredes*/ 
	adiciona(&p,px,py);
  }
  return p;
}


/* retorna 1 se o ponto estiver dentro do poligono
   ou 0 se estiver fora
 */
int ponto_em_poligono(int ax, int ay, poligono p)
{
  int count, px,py;
  float dist;
  count=interceptpoli(ax,ay,ax+20000,ay+22222,p,&px,&py,&dist);
  return (count%2);
  /*conta o numero de vezes que uma semi-reta originada em a cruza o poligono*/
  /*se for par, estah fora, se for impar, estah dentro.*/
}


/* exemplo de animacao para teste*/
void exemplo_animacao1()
{
  int i;
  poligono p=novo(BLUE,RED);
  
  for(i=0;i<5;i++) adiciona(&p,200+100*sin(i*6.0/5), 200+100*cos(i*6.0/5));
  
  setcolor(BLUE);
  rectangle(10,10,50,70);
  for(i=0; i<100; i++)
  {
    desenha(p);
	move(p,2,2);
	rotaciona(p,10);
	delay(500);
  }
}

/* exemplo de animacao para teste */
void exemplo_animacao2()
{
  int px,py,k,i;
  float dist;
  
  k=intercept(100,100,200,100,150,150,150,50,&px,&py,&dist);
  printf("%d %d %d %f\n",k,px,py,dist);
  
  cenario[0]=novo(BLUE,WHITE);
  adiciona(&cenario[0], 220,300);
  adiciona(&cenario[0], 240,300);
  adiciona(&cenario[0], 240,100);
  adiciona(&cenario[0], 220,100);
  n_cenario=1;

  for(i=10; i<=100; i+=10)
  {
    poligono p;
    p=explosao(200,200,i);
  
    desenha(cenario[0]);
    desenha(p);
  
    k=ponto_em_poligono(100,100,p);
    printf("%d ",k);
    k=ponto_em_poligono(220,220,p);
    printf("%d\n",k);
	
	libera(&p);
	delay(50);
  }
}

/* desenha_carrinho desenha o protagonista do jogo com centro em (x,y)
   olhando para a direcao dada pelo angulo em graus ang*/
void desenha_carrinho(int x, int y, float ang)
{
  poligono p1,p2,p3;
  p1=novo(YELLOW,GREEN);
  p2=novo(GREEN,YELLOW);
  p3=novo(GREEN,YELLOW);
  
  adiciona(&p1, -3,0); 
  adiciona(&p1, 3,0); 
  adiciona(&p1, 10,10); 
  adiciona(&p1, 0,20); 
  adiciona(&p1, -10,10); 

  adiciona(&p2, -9,3); 
  adiciona(&p2, -9,7); 
  adiciona(&p2, 9,7); 
  adiciona(&p2, 9,3);
  
  adiciona(&p3, -9,13); 
  adiciona(&p3, -9,17); 
  adiciona(&p3, 9,17); 
  adiciona(&p3, 9,13);
  
  move(p1,0,-10);
  move(p2,0,-10);
  move(p3,0,-10);
  
  escala(p1, 0.85, 1.0);
  escala(p2, 0.85, 1.0);
  escala(p3, 0.85, 1.0);

  rot(p1,90);
  rot(p2,90);
  rot(p3,90);
 
  rot(p1,ang);
  rot(p2,ang);
  rot(p3,ang);

  move(p1,x,y);
  move(p2,x,y);
  move(p3,x,y);
  
  desenha(p2);
  desenha(p3);
  desenha(p1);
  
  libera(&p1);
  libera(&p2);
  libera(&p3);
}
    

/* desenha_bomba desenha uma bomba colocada na posicao (x,y)
   com valor de timer dado por "timer" */  
void desenha_bomba(int x, int y, int timer)
{
  int i;
  poligono p1;
  p1=novo(LIGHTBLUE,CYAN);
  
  for(i=0;i<5;i++)
  {
    adiciona(&p1,x+10*sin(i*6.3/5), y+10*cos(i*6.3/5));
    adiciona(&p1,x+5*sin(i*6.3/5), y+5*cos(i*6.3/5));
  }
  
  rotaciona(p1,3.1416*2.0*timer/8.0);
  
  desenha(p1);
  libera(&p1);
}


int main()
{
  /*inicializa uma janela com duplo buffer  para a animacao nao ficar piscando*/ 
  initwindow(640,480,"Ces10bomber", 10, 10, true, true); 
  
  int k;
  
  float x=200, y=200, sp=2, s, c;
  int ang=0;
  
  float bombx, bomby, explox, exploy;
  int bombtimer=0, exploradius=0;
  
  /*construir elementos do cenario (paredes)*/
  n_cenario=3;
  
  cenario[0]=novo(BLUE,WHITE);
  cenario[1]=novo(BLUE,WHITE);
  cenario[2]=novo(BLUE,WHITE);
  
  adiciona(&cenario[0],500,20 );
  adiciona(&cenario[0],550,20 );
  adiciona(&cenario[0],650,100 );
  adiciona(&cenario[0],600,100 );

  adiciona(&cenario[1], 100,100 );
  adiciona(&cenario[1], 150,170 );
  adiciona(&cenario[1], 50, 120 );

  for(k=0;k<4;k++)
  {
    adiciona(&cenario[2],400+100*sin(k*6.3/4), 300+100*cos(k*6.3/4));
    adiciona(&cenario[2],400+50*sin(k*6.3/4), 300+50*cos(k*6.3/4));
  }
  
  /* malha principal do jogo */
  while(1) /* repete ateh um break ou exit*/
  {
    float newx,newy;
	int moveok;
	int i;
	
	clearviewport(); /*apaga janela*/
    c=cos(-ang*3.1416/180.0);
	s=sin(-ang*3.1416/180.0);
	while(kbhit()) /*enquanto tem teclas na fila*/
	{
	  k=getch();
	  switch(k)
	  {
	     case KEY_UP: /*seta para cima -> anda para frente*/
		   newx=x+sp*c;
		   newy=y+sp*s;
		   moveok=1;
		   for(i=0; i<n_cenario; i++)
		   { /* teste para carrinho nao andar para dentro da parede*/
		     if(ponto_em_poligono(newx,newy,cenario[i])) moveok=0;
		   }
		   if(moveok)
		   {
		     x=newx;
			 y=newy;
		   }
		   break;
		 case KEY_RIGHT: /*vira a direita*/
		   ang=(ang+360-5)%360;
		   break;
		 case KEY_LEFT: /*vira a esquerda*/
		   ang=(ang+5)%360;
		   break;
		 case 32: /*espaco -> deixa uma bomba*/
		   if(bombtimer==0) /*limitado a apenas uma bomba por vez*/
		   {
		     bombx=x;
		     bomby=y;
		     bombtimer=200; /*bomba explode apos 200 tu*/
		   }
		   break;
		 case 27: /*tecla esca sai do jogo*/
		   exit(0);
	  }
	}
	if(bombtimer) /*se bomba esta acionada*/
	{
	  bombtimer--; /*decrementa relogio da bomba*/
	  if(bombtimer==0) /*se hora de explodir*/
	  {
	    exploradius=10; /*seta raio da explosao para 10 pixels*/
		explox=bombx; /*posicao da explosao eh a posicao da bomba*/
		exploy=bomby;
	  }
	  desenha_bomba(bombx,bomby,bombtimer);
	}
	if(exploradius) /*se explosao ocorrendo*/
	{
	  poligono expl; /*cria poligono explosao*/
	  expl=explosao(explox,exploy,exploradius);
	  if(ponto_em_poligono(x,y,expl)) /*testa se carrinho dentro da explosao*/
	  {
	    x=0;y=0; /*explode carrinho (jogo nao-violento, soh move para 0,0)*/
	  }
	  desenha(expl);
	  libera(&expl); /*libera memoria do poligono explosao*/
	  exploradius+=10; /*faz explosao crescer*/
	  if(exploradius>100) exploradius=0; /*quando raio chega em 100 acaba expl*/
	}  
    desenha_carrinho(x,y,ang);
	for(k=0; k<n_cenario; k++) desenha(cenario[k]);
	swapbuffers(); /*mostra tudo o que foi desenhado, trocando o buffer de tela*/
    delay(30); /*espera 30 milisecs*/
  }
	
  getchar();
  closegraph();
  
}

