/* poligono.c
 * biblioteca para lidar com poligonos, onde vertices sao guardados como
 * uma lista ligada.
 *  
 * Curso CES-10
 * Autor: Carlos Henrique Q. Forster
 * Instituto Tecnolgico de Aeronutica
 * 02/06/2008
 */

#include "poligono.h"
#include <graphics.h>
#include <math.h>


/* novo retorna um poligono com zero vertices com cores de preenchimento
   e contorno definidas */
poligono novo(int preenchimento, int contorno)
{
  poligono p;
  p.vert=NULL;
  p.preenchimento=preenchimento;
  p.contorno=contorno;
  p.nvertices=0;
  return p;
}

/*buffer para desenhar os poligonos, acesso deve ser privado*/
int polipoints[1000];

/* desenha o poligono na janela grafica - interior e contorno */
void desenha(poligono p)
{
  int i=0;
  vertice *v=p.vert;
  while(v)
  {
    polipoints[i++]=v->x;
	polipoints[i++]=v->y;
	v=v->prox;
  }
  setfillstyle(SOLID_FILL,p.preenchimento);
  fillpoly(p.nvertices,polipoints);
  
  polipoints[i++]=p.vert->x;
  polipoints[i++]=p.vert->y;
  setcolor(p.contorno);
  drawpoly(p.nvertices+1,polipoints);
}


/* destaca os vertices de um poligono com quadrados */
void desenha_pontos(poligono p)
{
  vertice *v=p.vert;
  while(v)
  {
    setcolor(WHITE);
    rectangle(v->x-5,v->y-5,v->x+5,v->y+5);
	v=v->prox;
  }
}   

/* adiciona vertices a um poligono ja criado */
void adiciona(poligono *p, int x, int y)
{
  vertice *v=(vertice *) malloc(sizeof(vertice));
  v->x=x;
  v->y=y;
  v->prox=p->vert;
  p->vert=v;
  p->nvertices++;
}
  

/* desloca todos os vertices de um poligono pelo vetor (dx,dy) */  
void move(poligono p, int dx, int dy)
{
  vertice *v=p.vert;
  while(v)
  {
    v->x+=dx;
	v->y+=dy;
	v=v->prox;
  }
}


/* multiplica todos os vertices por um valor*/
void escala(poligono p, double sx, double sy)
{
  vertice *v=p.vert;
  while(v)
  {
    v->x=v->x*sx;  
    v->y=v->y*sy;  
    v=v->prox;
  }
}


/* copia vertice por vertice criando novo poligono*/
void copia(poligono *q, poligono p)
{
  vertice *v=p.vert;
  *q=novo(p.preenchimento,p.contorno);
  while(v)
  {
    adiciona(q,v->x,v->y);
	v=v->prox;
  }
}


/* rotaciona poligono em relacao a origem (0,0) */
void rot(poligono p, double graus)
{
  vertice *v=p.vert;
  while(v)
  {
    int x,y;
	float s,c;
	s=sin(graus*3.1416/180.0);
	c=cos(graus*3.1416/180.0);
	x=v->x;
	y=v->y;
	v->x=x*c+y*s;
	v->y=-x*s+y*c;
	v=v->prox;
  }
}


/* rotaciona poligono em relacao ao centro do retangulo que o envolve*/
void rotaciona(poligono p, double graus)
{
  int pivox,pivoy;
  pivox=(maxx(p)+minx(p))/2;
  pivoy=(maxy(p)+miny(p))/2;
  move(p,-pivox,-pivoy);
  rot(p,graus);
  move(p,pivox,pivoy);
}

/* as proximas funcoes determinam as coordenadas de um retangulo que envolve
   um poligono dado poligono: maxx, minx, maxy, miny
*/
int maxx(poligono p)
{
  vertice *v=p.vert;
  int max;
  max=p.vert->x;
  while(v)
  {
    if(v->x > max) max=v->x;
	v=v->prox;
  }
  return max;
}

int maxy(poligono p)
{
  vertice *v=p.vert;
  int max;
  max=p.vert->y;
  while(v)
  {
    if(v->y > max) max=v->y;
	v=v->prox;
  }
  return max;
}

int minx(poligono p)
{
  vertice *v=p.vert;
  int min;
  min=p.vert->x;
  while(v)
  {
    if(v->x < min) min=v->x;
	v=v->prox;
  }
  return min;
}

int miny(poligono p)
{
  vertice *v=p.vert;
  int min;
  min=p.vert->y;
  while(v)
  {
    if(v->y < min) min=v->y;
	v=v->prox;
  }
  return min;
}

/* libera a memoria dinamica associada a um dado poligono*/
void libera(poligono *p)
{
  vertice *v=p->vert, *q;
  while(v)
  {
	q=v;
	v=v->prox;
	free(q);
  }
  p->vert=NULL;
}

